#include <saplib/win/all.h>
#pragma hdrstop

#define NTAB 10
#define TIMER1 50
#define LINEW 36
#define SPACEW 8
#define CRINFOW 50
#define SCRW 440
#define SCRH (BW+(LINEW+SPACEW)*NTAB+(BW-SPACEW)+CRINFOW)

struct rec_info
{
	char		name[40];
	int		seconds;
	char		date[40];
};

static rec_info  g_tab[NTAB];
static rec_info  g_initab[NTAB]=
{
	{"slot-1",	200,	"-"},
	{"slot-2",	220,	"-"},
	{"slot-3",	240,	"-"},
	{"slot-4",	300,	"-"},
	{"slot-5",	350,	"-"},
	{"slot-6",	400,	"-"},
	{"slot-7",	450,	"-"},
	{"slot-8",	500,	"-"},
	{"slot-9",	600,	"-"},
	{"slot-10",	650,	"-"},
};

static int g_lasti=BADI;

int		TAB_GetPlace(int seconds);
bool		TAB_WritePlayer(const char*name, int seconds);
void		TAB_ResetPlayer(int);
void		TAB_Reset();


class VScores: public VMsgWindow
{
	HBITMAP		m_back;
	HFONT		m_font1;
	void		CmdClose();
protected:
	LRESULT		MsgCreate();
	void		MsgDestroy();
	void		MsgPaint(VDC&O,RECT win);
	void		MsgSize(SIZE);
	BOOL		MsgKeyDown(BYTE, int);
	void		MsgTimer(UINT);
	void		MsgMouse(UINT msg, POINT pt);
public:
	VScores();
       ~VScores();
};

void TAB_Reset()
{
	memcpy(g_tab, g_initab, sizeof(g_tab));
}

void TAB_ResetPlayer(int i)
{
	memcpy(&g_tab[i], &g_initab[i], sizeof(g_tab[0]));
}


/*=============
  Note: we overwrite an already existing line,
  and delete the previous player from the table!
=============*/
bool TAB_WritePlayer(const char*name, int seconds)
{
	int		place;
	char		date[80];
	SYSTEMTIME	now;

	GetLocalTime(&now);

	GetDateFormat(0, 0, &now, "yyyy dd MMMM',' dddd", date, sizeof(date));

	if((place=TAB_GetPlace(seconds))  != BADI)
	{
		strcpy(g_tab[place].name, name);
		strcpy(g_tab[place].date, date);
		g_tab[place].seconds=seconds;
		g_lasti=place;
		return 1;
	}

	return 0;
}

/*=============
   Returns: a zero base index in the scores table or -1.
=============*/
int TAB_GetPlace(int seconds)
{
	for(int i=0; i<LEN(g_tab); i++)
	{
		if(seconds <= g_tab[i].seconds)
			return i;
	}

	return BADI;
}


/*=============
   Scores view window
=============*/

VScores::VScores()
{}

VScores::~VScores()
{}

LRESULT VScores::MsgCreate()
{
	m_back=LoadBitmap(ThisModule(), "back");
	CheckError(!m_back);
	m_font1=create_font(10,24,FW_BOLD, "Arial");
	CheckError(!m_font1);

	EnableWindow(GetParent(), FALSE);
	MoveTo(SCRW,SCRH);
	SetFocus(hthis);
	SetTimer(hthis,TIMER1,TIMER1,NULL);
	return 0;
}

void VScores::CmdClose()
{
	EnableWindow(GetParent(), TRUE);
	DestroyWindow(hthis);
}

void VScores::MsgDestroy()
{
	DeleteObject(m_back);
	DeleteObject(m_font1);
}

BOOL VScores::MsgKeyDown(BYTE, int vKey)
{
	switch(vKey)
	{
	case VK_ESCAPE:
		CmdClose();
		break;
	default:
		return FALSE;
	}
	return TRUE;
}

void VScores::MsgSize(SIZE)
{}

void VScores::MsgMouse(UINT msg, POINT pt)
{
static int pressed=0;
static WINDOWPLACEMENT wp;
static POINT start;

	switch(msg)
	{
	case WM_RBUTTONDOWN:
		CmdClose();
		break;

	case WM_LBUTTONDOWN:
		SetCapture(hthis);
		wp.length=sizeof(wp);
		GetWindowPlacement(hthis, &wp);
		pressed=3;
		start=pt;
		break;

	case WM_MOUSEMOVE:
		if(pressed==3)
		{
			OffsetRect(&wp.rcNormalPosition, pt.x-start.x, pt.y-start.y);
			SetWindowPlacement(hthis, &wp);
		}
		break;

	case WM_LBUTTONUP:
		if(pressed==3)
		{
			pressed=0;
			ReleaseCapture();
		}
		break;

	}
}

void VScores::MsgTimer(UINT id)
{
	if(id==TIMER1)
	{
		Invalidate();
		return;
	}
}

void VScores::MsgPaint(VDC&P,RECT win)
{
	VDC O;

	O.MemoryDC(P.hthis,win);

				// *) draw background & border
	RECT rc=win;
	O.Fill(rc,m_back);
	O.NoBrush();
	O.NoPaper();

	for(int i=0; i<5;i++)
	{
		InflateRect(&rc, -1,-1);
		O.Pen(RGB(0,0,0), 1);
		O.Rect(rc);
		InflateRect(&rc, -1,-1);
	}

				// *) draw the high scores
	SIZE	sz;
	int	ypos=BW;
	vstring str;
				// *) draw the other scores
	int time=(GetTickCount()/TIMER1);

	for(int i=0; i<NTAB;i++)
	{
		int percent=50+abs(25-(time+MulDiv(i,50,NTAB))%50);

		COLORREF cLine=percent_of_color(RGB(95,255,255), percent);
		COLORREF cText=RGB(0,0,0);

		str.format("%i", g_tab[i].seconds);

		int	xpos=BW+BW;

		SetRect(&rc, BW, ypos, SCRW-BW, ypos+LINEW);
		O.Brush(cLine);
		O.Pen(cText,1);
		O.Rect(rc);
		DrawEdge(O.hthis, &rc, EDGE_SUNKEN, BF_RECT);

		O.Select(m_font1);
		O.Text(cText);
		O.GetSize(g_tab[i].name, &sz);
		O.Text(xpos,ypos+(LINEW-sz.cy)/2,g_tab[i].name);
		O.Select(GetStockObject(ANSI_VAR_FONT));
		O.Text(cText);
		O.GetSize(str, &sz);
		O.Text(xpos+=180,ypos+(LINEW-sz.cy)/2,str);

		O.Text((i==g_lasti) ? RGB(0,0,255):cText);
		O.Select(GetStockObject(DEFAULT_GUI_FONT));
		O.GetSize(g_tab[i].date, &sz);
		O.Text(xpos+=40,ypos+(LINEW-sz.cy)/2,g_tab[i].date);

		ypos+=LINEW+SPACEW;
	}

	static char verinfo[]=
			"\"Floating Squares\"\r\n"
			"Created by Zint, Russia\r\n"
			"Site: zint.narod.ru,  e-mail: Zint@yandex.ru\r\n";

	O.Select(GetStockObject(DEFAULT_GUI_FONT));
	O.Text(RGB(0,0,0));

	SetRect(&rc, 0, ypos, SCRW, SCRH);

	RECT rcText=rc;
	DrawText(O.hthis,verinfo,strlen(verinfo),&rcText,DT_CENTER|DT_CALCRECT);
	InflateRect(&rc,
		-(RCW(rc)-RCW(rcText))/2,
		-(RCH(rc)-RCH(rcText))/2);

	DrawText(O.hthis,verinfo,strlen(verinfo),&rc,DT_CENTER);

	O.FlushMemoryDC();
}

