#include <saplib/win/all.h>
#pragma hdrstop

#define DEBUG 0
#define USEDEFTABLE 0

#define BW 20
#define CX 9
#define CY 9
#define CELLW 52
#define CELLH 52
#define SIGNW 50
#define STATUSH 0
#define MENUH 48
#define BUTTONW 100
#define BUTTONH 26
#define BUTTONY (FH-MENUH)
#define DIGITW 12
#define DIGITH 20
#define TARGETW 40
#define FW (BW+CELLW*CX+(2*SPACEW) + BW)
#define FH (STATUSH + BW+CELLH*CY+BW + (2*SPACEW) + MENUH)
#define SND_Move "move"
#define SND_Press "press"
#define SND_Win "win"
#define ROOT_KEY "SOFTWARE\\FloatingSquares"
#define VALUE_SCORES "Scores table"
#define VALUE_USAGE "Run counter"
#define GAME_TIMER 300


class VApp: public VMsgWindow
{
	struct s_button
	{
		int	x;
		int	y;
		const char* text;
	};

	static s_button buttons[];

	int		m_tab[CX*CY];
	int		m_origin[CX*CY];
	POINT		m_coord[CX*CY];
	HBITMAP		m_figures,
			m_buttons,
			m_back,
			m_digits,
			m_target;
	HFONT		m_font1;
	HCURSOR		m_hand1,
			m_hand2;
	int		m_btnst[3];
	int		m_starttime;

	bool		GetLogicalCoords(POINT pos, int* px, int *py)const;
	bool		TraceButton(int&, POINT pt,int status);
	void		Draw(VDC&, const RECT&);
	void		MoveUp(int x, int num);
	void		MoveLeft(int y, int num);
	void		FormatSignsPos();
	void		MoveSignsPos(POINT &start, POINT pos);
	void		NewGame();
	void		ShowScores();
	void		UpdateScoresTable();
	void		CloseApp();
	bool		LoadConfiguration();
	bool		SaveConfiguration();
protected:
	LRESULT		MsgCreate();
	void		MsgDestroy();
	void		MsgPaint(VDC&O,RECT win);
	void		MsgSize(SIZE);
	BOOL		MsgKeyDown(BYTE, int);
	void		MsgTimer(UINT);
	void		MsgMouse(UINT msg, POINT pt);
public:
	VApp();
       ~VApp();
};


#include "table.cpp"


VApp::s_button VApp::buttons[]=
{
	{BW+130, BUTTONY, "new game"},
	{BW+130+(BUTTONW+10), BUTTONY, "info"},
	{BW+130+(BUTTONW+10)*2, BUTTONY, "close"},
};

static int g_signtab[]={
	1,1,1,2,2,2,3,3,3,
	1,1,1,2,2,2,3,3,3,
	1,1,1,2,2,2,3,3,3,
	4,4,4,5,5,5,6,6,6,
	4,4,4,5,5,5,6,6,6,
	4,4,4,5,5,5,6,6,6,
	7,7,7,8,8,8,9,9,9,
	7,7,7,8,8,8,9,9,9,
	7,7,7,8,8,8,9,9,9,
};

/*=========
   0 [1] 2 3 [4] 5 6 [7] 9
=========*/

static int g_movable[]=
{
	1,0,1,1,0,1,1,0,1
};


// Shows how much the user ran this application
static int g_usage=0;


static void SUB_XorData(char*buf, size_t size)
{
	int c=size;
	for(int i=0; i<size; i++)
	{
		c+=(848564183+i*i) >> 1;
		buf[i]^=(char)c;
	}
}

static void SUB_Sound(const char *name)
{
	HRSRC hr;
	HGLOBAL hg;
	PVOID hw;

	if(
		(hr=FindResource(ThisModule(), name, "wave"))!=NULL &&
		(hg=LoadResource(ThisModule(), hr))!=NULL &&
		(hw=LockResource(hg))!=NULL)
	{
		sndPlaySound((char*)hw, SND_ASYNC|SND_MEMORY);
	}
}


int APIENTRY WinMain(HINSTANCE,HINSTANCE,LPSTR,int)
{
	VApp box;
	box.Create(WS_POPUP);
	box.Process();
	return 0;
}

VApp::VApp()
{
	memset(m_btnst, 0, sizeof(m_btnst));
}

VApp::~VApp()
{}

LRESULT VApp::MsgCreate()
{
	randomize();
	TAB_Reset();
	SetWindowText(hthis, "Floating Squares");
#if !USEDEFTABLE
	LoadConfiguration();
#endif

#if USEDEFTABLE
	TAB_WritePlayer("Angela", 201);
	TAB_WritePlayer("John", 208);
	TAB_WritePlayer("Smanata", 223);
	TAB_WritePlayer("DF:-)", 265);
	TAB_WritePlayer("Slavik", 329);

	TAB_WritePlayer("DF", 355);
	TAB_WritePlayer("Master Vic", 412);
	TAB_WritePlayer("Samata", 446);
	TAB_WritePlayer("Victor", 470);
	TAB_WritePlayer("John", 477);
#endif
	m_figures=LoadBitmap(ThisModule(), "images");
	m_buttons=LoadBitmap(ThisModule(), "buttons");
	m_back=LoadBitmap(ThisModule(), "back");
	m_digits=LoadBitmap(ThisModule(), "digits");
	m_target=LoadBitmap(ThisModule(), "target");
	m_hand1=LoadCursor(ThisModule(), "hand1");
	m_hand2=LoadCursor(ThisModule(), "hand2");
	m_font1=create_font(10,24,FW_BOLD, "Arial");

	SetCursor(m_hand1);

	NewGame();
	MoveTo(FW, FH);
	SetFocus(hthis);
	SetTimer(hthis, GAME_TIMER, GAME_TIMER, NULL);
	return 0;
}

void VApp::MsgDestroy()
{
	DeleteObject(m_hand1);
	DeleteObject(m_hand2);
	DeleteObject(m_figures);
	DeleteObject(m_buttons);
	DeleteObject(m_back);
	DeleteObject(m_digits);
	DeleteObject(m_target);
	DeleteObject(m_font1);
#if 1
	PostQuitMessage(0);
#endif
}

void VApp::CloseApp()
{
#if !USEDEFTABLE
	SaveConfiguration();
#endif
	DestroyWindow(hthis);
}

BOOL VApp::MsgKeyDown(BYTE, int vKey)
{
	static char LastKeys[21]={0};

	if((vKey>='A' && vKey<='Z') || vKey==' ')
	{
		memmove(LastKeys+1, LastKeys, 19);
		LastKeys[0]=(char)vKey;

		// END: format the signs
		if(strncmp(LastKeys, "DNE", 3)==0)
		{
			SUB_Sound(SND_Press);
			memcpy(m_tab, g_signtab, sizeof(m_tab));
			Invalidate();
			UpdateWindow(hthis);
			return TRUE;
		}

		// EXIT: exit the application without saving the config file
		if(strncmp(LastKeys, "TIXE", 4)==0)
		{
			SUB_Sound(SND_Press);
			DestroyWindow(hthis);
			return TRUE;
		}

		// RESET
		if(strncmp(LastKeys, "TESER", 5)==0)
		{
			SUB_Sound(SND_Press);
			TAB_Reset();
			return TRUE;
		}

		// TAB
		if(strncmp(LastKeys, "BAT", 3)==0)
		{
			SUB_Sound(SND_Press);
			ShowScores();
			return TRUE;
		}
	}

	switch(vKey)
	{
	case VK_ESCAPE:
		CloseApp();
		break;
	default:
		return FALSE;
	}
	return TRUE;
}

void VApp::MsgSize(SIZE)
{}

void VApp::MsgPaint(VDC& P,RECT win)
{
	VDC O;
	O.MemoryDC(P.hthis, win);
	Draw(O, win);
	O.FlushMemoryDC();
}

/* Registry IO
*/
bool VApp::SaveConfiguration()
{
	HKEY		hkey;
	DWORD		dwDisp;
	LONG		err;
	char		buf[4000];

	err=RegCreateKeyEx(
		HKEY_LOCAL_MACHINE,
		ROOT_KEY,
		0,		// no options
		NULL,		// class
		0,		// non volatile
		KEY_ALL_ACCESS,
		NULL,		// no security attr
		&hkey,
		&dwDisp);

	if(err != ERROR_SUCCESS)
		return 0;

	// save scores table
	char*	p=buf;
	for(int i=0; i<NTAB; i++)
	{
		strcpy(p, g_tab[i].name);
		p+=strlen(p)+1;
		strcpy(p, g_tab[i].date);
		p+=strlen(p)+1;
		memcpy(p, &g_tab[i].seconds, sizeof(int));
		p+=sizeof(int);
	}

	size_t	dwSize=p-buf;

	SUB_XorData(buf, dwSize);

	RegSetValueEx(hkey, VALUE_SCORES, 0, REG_BINARY, (PBYTE)buf, dwSize);

	// save usage counter
	RegSetValueEx(hkey, VALUE_USAGE, 0, REG_DWORD, (PBYTE)&g_usage, sizeof(g_usage));

	RegCloseKey(hkey);
	return 1;
}

bool VApp::LoadConfiguration()
{
	HKEY		hkey;
	DWORD		dwDisp;
	DWORD		dwType;
	DWORD		dwSize;
	LONG		err;
	char		buf[4000];

	err=RegCreateKeyEx(
		HKEY_LOCAL_MACHINE,
		ROOT_KEY,
		0,		// no options
		NULL,		// class
		0,		// non volatile
		KEY_ALL_ACCESS,
		NULL,		// no security attr
		&hkey,
		&dwDisp);

	if(err != ERROR_SUCCESS)
		return 0;

	// load usage counter
	dwSize=sizeof(g_usage);
	err=RegQueryValueEx(hkey, VALUE_USAGE, NULL, &dwType, (PBYTE)&g_usage, &dwSize);
	if(err == ERROR_SUCCESS && dwType==REG_DWORD)
	{
		g_usage++;
	}

	// load table
	dwSize=sizeof(buf);
	err=RegQueryValueEx(hkey, VALUE_SCORES, NULL, &dwType, (PBYTE)buf, &dwSize);
	if(err == ERROR_SUCCESS && dwType==REG_BINARY)
	{
		SUB_XorData(buf, dwSize);
		char*	p=buf;
		for(int i=0; i<NTAB; i++)
		{
			strcpy(g_tab[i].name, p);
			p+=strlen(p)+1;
			strcpy(g_tab[i].date, p);
			p+=strlen(p)+1;
			memcpy(&g_tab[i].seconds, p, sizeof(int));
			p+=sizeof(int);
		}
	}

	RegCloseKey(hkey);
	return 1;
}

void VApp::MoveUp(int x, int num)
{
	int	tmp[CY];

	CheckError(x<0||x>=CX);

	num %= CY;
	while(num<0) num += CY;

	for(int i=CY; --i>=0;)
		tmp[i]=m_tab[CX*i+x];

	for(int i=CY; --i>=0;)
		m_tab[CX*((i+num)%CY)+x]=tmp[i];
}

void VApp::MoveLeft(int y, int num)
{
	int	tmp[CX];

	CheckError(y<0||y>=CY);

	num%=CX;
	if(num<0) num=CX+num;

	for(int i=CX; --i>=0;)
		tmp[i]=m_tab[CX*y+i];

	for(int i=CX; --i>=0;)
		m_tab[CX*y+(i+num)%CX]=tmp[i];
}

void VApp::FormatSignsPos()
{
	for(int x=CX; --x>=0;)
	{
		for(int y=CY; --y>=0;)
		{
			int i=y*CX+x;

			m_coord[i].x=BW + (x*CELLW) + (x/3*SPACEW);
			m_coord[i].y=STATUSH+BW + (y*CELLH) + (y/3*SPACEW);
		}
	}
}

void VApp::ShowScores()
{
	VScores box;
	box.Create(WS_POPUP,hthis);
	box.Process();
}

void VApp::NewGame()
{
	memcpy(m_tab, g_signtab, sizeof(m_tab));
	FormatSignsPos();

				// *) shuffle the field
	for(int i=100; --i>=0;)
	{
		for(int x=CX; --x>=0;)
			if(g_movable[x]) MoveUp(x, rand()%CY);

		for(int y=CY; --y>=0;)
			if(g_movable[y]) MoveLeft(y, rand()%CX);
	}

	m_starttime=GetTickCount();
}

void VApp::UpdateScoresTable()
{
		// take the current time
	int seconds=(GetTickCount()-m_starttime)/1000;

	SUB_Sound(SND_Win);

	vstring str;

	if(TAB_GetPlace(seconds) != BADI && q_string(hthis, &str,
		"You are in high scores table",
		"Enter your name:"))
	{
		TAB_WritePlayer(String(str), seconds);
		ShowScores();
	}
}

bool VApp::GetLogicalCoords(POINT pos, int* px, int *py)const
{
	POINT	tmp;

	for(int x=0; x<CX; x++)
	{
		for(int y=0; y<CY; y++)
		{
			tmp=m_coord[CX*y+x];

			if(pos.x>=tmp.x && pos.x<tmp.x+CELLW)
			if(pos.y>=tmp.y && pos.y<tmp.y+CELLH)
			{
				*px=x;
				*py=y;
				return 1;
			}
		}
	}
	return 0;
}

void VApp::MsgTimer(UINT id)
{
	if(id==GAME_TIMER)
	{
		Invalidate();
		return;
	}
}

void VApp::Draw(VDC&O, const RECT&dest)
{
	RECT rc=dest;
				// *) draw the backgound and the border
	O.Fill(rc,m_back);
	O.NoBrush();
	O.Pen(RGB(0,0,0), 1);
	for(int i=0; i<5;i++)
	{
		InflateRect(&rc, -1,-1);
		O.Rect(rc);
		InflateRect(&rc, -1,-1);
	}

	VDC T;
	T.CompatibleDC(O.hthis);

				// *) draw the status line
	int cx=BW;
	int cy=FH-MENUH-BW/2;
	T.Select(m_target);
	O.BitCopy(cx, cy, TARGETW, TARGETW, T, 0, 0);

	SetRect(&rc, cx,cy,cx+TARGETW,cy+TARGETW);
	O.NoBrush();
	O.Pen(RGB(0,0,0));
	O.Rect(rc);
	InflateRect(&rc, 2,2);
	O.Rect(rc);

				// *) draw the timer
	vstring digs;
	int xpos=BW+40+BW;
	int ypos=FH-MENUH-BW/2;

	O.NoBrush();
	O.Pen(RGB(0,0,0));
	SetRect(&rc, xpos,ypos,xpos+4*DIGITW,ypos+DIGITH);
	InflateRect(&rc, 2,2);
	O.Rect(rc);

	T.Select(m_digits);
	digs.format("%04i", (GetTickCount()-m_starttime)/1000);
	for(int i=0; digs[i]; i++)
	{
		O.BitCopy(xpos, ypos, DIGITW, DIGITH,
			T, DIGITW*(digs[i]-'0'), 0);
		xpos+=DIGITW;
	}

			// *) draw squares
	T.Select(m_figures);
	for(int x=0; x<CX; x++)
	{
		for(int y=0; y<CY; y++)
		{
			int i=y*CX+x;
			int sign=m_tab[i]-1;

			xpos=m_coord[i].x + (CELLW-SIGNW)/2;
			ypos=m_coord[i].y + (CELLH-SIGNW)/2;

			O.BitCopy(xpos, ypos, SIGNW, SIGNW,
				T, sign*SIGNW, 0);

			SetRect(&rc, xpos-1, ypos-1, xpos+CELLW-2, ypos+CELLH-2);

			// if(!g_movable[x] && !g_movable[y])
			DrawEdge(O.hthis, &rc, EDGE_RAISED, BF_RECT);
		}
	}

				// *) draw menu buttons
	T.Select(m_buttons);
	O.NoPaper();
	O.Text(RGB(255,255,255));
	O.Select(GetStockObject(DEFAULT_GUI_FONT));
	for(int i=0; i<3; i++)
	{
		SIZE sz;
		O.GetSize(buttons[i].text, &sz);

		int x=buttons[i].x+(BUTTONW-sz.cx)/2;
		int y=buttons[i].y+(BUTTONH-sz.cy)/2;

		switch(m_btnst[i])
		{
		case 0:
			O.BitCopy(buttons[i].x, buttons[i].y,
				BUTTONW, BUTTONH, T, BUTTONW, 0);

			O.Text(RGB(255,255,255));
			O.Text(x, y, buttons[i].text);
			break;
		case 1:
			O.BitCopy(buttons[i].x, buttons[i].y,
				BUTTONW, BUTTONH, T, BUTTONW, 0);

			O.Text(RGB(0,255,255));
			O.Text(x, y, buttons[i].text);
			break;
		case 2:
			O.BitCopy(buttons[i].x, buttons[i].y,
				BUTTONW, BUTTONH, T, 0, 0);

			O.Text(RGB(225,225,225));
			O.Text(x+2, y+2, buttons[i].text);
			break;
		}
	}
}

bool VApp::TraceButton(int& btn, POINT pt, int status)
{
	int i;

	for(i=3; --i>=0;)
	{
		if(pt.x >= buttons[i].x && pt.x < buttons[i].x+BUTTONW)
		if(pt.y >= buttons[i].y && pt.y < buttons[i].y+BUTTONH)
			break;
	}

	if(i != btn || (i>=0 && m_btnst[i]!=status))
	{
		if(btn>=0)
		{
			m_btnst[btn]=0;
		}
		btn=i;
		if(btn>=0)
		{
			m_btnst[btn]=status;

			if(status==2)
				SUB_Sound(SND_Press);
		}
		Invalidate();
		UpdateWindow(hthis);
	}
	return btn>=0;
}

/*
  Defines when to switch squares from one 3x3 zone to other.
*/
#pragma argsused
static void SUB_CalcPos(int pos, int oppos, int W, int *proll, int *pshift)
{
#if 0
	int delta=abs(pos)-abs(oppos);
	int tmp=(delta * 25/(abs(pos)-delta+25)) * (pos/abs(pos));
#else
	int tmp=pos;
#endif
	if(pos>0)
	{
		*proll=(tmp+W/2)/W;
		*pshift=MulDiv((tmp+W/2)%W-W/2, 70,70);
	} else
	{
		*proll=(tmp-W/2)/W;
		*pshift=MulDiv((tmp-W/2)%W+W/2, 70,70);
	}
}

void VApp::MoveSignsPos(POINT& start, POINT pos)
{
	int lx, ly;
	int roll;
	int shift;

	FormatSignsPos();

	memcpy(m_tab, m_origin, sizeof(m_tab));

	if(!GetLogicalCoords(start, &lx, &ly))
	{
		start=pos;
		return;
	}

	int cx=pos.x - start.x;
	int cy=pos.y - start.y;

//	if(abs(cx) > abs(cy))
	if(g_movable[ly])
	{
		SUB_CalcPos(cx, cy, CELLW, &roll, &shift);

		MoveLeft(ly, roll);
		for(int x=CX; --x>=0;)
		{
			m_coord[ly*CX+x].x+=shift;
		}

		if(roll)
		{
			memcpy(m_origin, m_tab, sizeof(m_origin));
			SUB_Sound(SND_Move);
			start.x+=roll*CELLW;
		}
	}

//	else if(abs(cy) > abs(cx))
	if(g_movable[lx])
	{
		SUB_CalcPos(cy, cx, CELLH, &roll, &shift);

		MoveUp(lx, roll);

		for(int y=CY; --y>=0;)
		{
			m_coord[y*CX+lx].y+=shift;
		}

		if(roll)
		{
			memcpy(m_origin, m_tab, sizeof(m_origin));
			SUB_Sound(SND_Move);
			start.y+=roll*CELLH;
		}
	}
}

void VApp::MsgMouse(UINT msg, POINT pt)
{
	enum {NOMOVE, MOVE_POS, MOVE_SIGN, MOVE_PRESS};

	static int pressed=NOMOVE;
	static WINDOWPLACEMENT wp;
	static POINT start;
	static int btn=-1;


	int lx, ly;
	RECT rc;

	SetRect(&rc, BW, BW, FW-BW, FH-BW-STATUSH);

	switch(msg)
	{
	case WM_LBUTTONDOWN:
		SetCapture(hthis);
		::SetCursor(m_hand2);

		if(GetLogicalCoords(pt, &lx, &ly))
		{
			pressed=MOVE_SIGN;
			start=pt;
			memcpy(m_origin, m_tab, sizeof(m_origin));
		}
		else if(TraceButton(btn, pt, 2))
		{
			pressed=MOVE_PRESS;
		}
		else if(!PtInRect(&rc, pt))
		{
			pressed=MOVE_POS;
			start=pt;
			wp.length=sizeof(wp);
			GetWindowPlacement(hthis, &wp);
		}
		break;

	case WM_MOUSEMOVE:
		if(!pressed)
		{
			TraceButton(btn,pt, 1);
			break;
		}

		::SetCursor(m_hand2);

		switch(pressed)
		{
		case MOVE_SIGN:
			MoveSignsPos(start, pt);
			Invalidate();
			UpdateWindow(hthis);
			return;

		case MOVE_POS:
			OffsetRect(&wp.rcNormalPosition, pt.x-start.x, pt.y-start.y);
			SetWindowPlacement(hthis, &wp);
			return;

		case MOVE_PRESS:
			TraceButton(btn,pt, 2);
			return;

		}
		break;

	case WM_LBUTTONUP:
		ReleaseCapture();

		switch(pressed)
		{
		case MOVE_SIGN:
			pressed=NOMOVE;
			FormatSignsPos();
			Invalidate();
			UpdateWindow(hthis);

			if(memcmp(m_tab, g_signtab, sizeof(m_tab)) == 0)
			{
				UpdateScoresTable();
				NewGame();
				break;
			}
			return;

		case MOVE_POS:
			pressed=NOMOVE;
			return;

		case MOVE_PRESS:
			pressed=NOMOVE;
			TraceButton(btn, pt, 0);

			switch(btn)
			{
			case 0:
				NewGame();
				break;

			case 1:
				ShowScores();
				break;

			case 2:
				CloseApp();
				break;
			}
			return;
		}

		break;
	}
}
